/* ############################# MPC License ############################## */
/* # Wed Nov 19 15:19:19 CET 2008                                         # */
/* # Copyright or (C) or Copr. Commissariat a l'Energie Atomique          # */
/* #                                                                      # */
/* # IDDN.FR.001.230040.000.S.P.2007.000.10000                            # */
/* # This file is part of the MPC Runtime.                                # */
/* #                                                                      # */
/* # This software is governed by the CeCILL-C license under French law   # */
/* # and abiding by the rules of distribution of free software.  You can  # */
/* # use, modify and/ or redistribute the software under the terms of     # */
/* # the CeCILL-C license as circulated by CEA, CNRS and INRIA at the     # */
/* # following URL http://www.cecill.info.                                # */
/* #                                                                      # */
/* # The fact that you are presently reading this means that you have     # */
/* # had knowledge of the CeCILL-C license and that you accept its        # */
/* # terms.                                                               # */
/* #                                                                      # */
/* # Authors:                                                             # */
/* #   - PERACHE Marc marc.perache@cea.fr                                 # */
/* #                                                                      # */
/* ######################################################################## */
#include <stdlib.h>
#include <stdio.h>
#include <sched.h>
#include <string.h>
#include "sctk_low_level_comm.h"
#include "sctk_debug.h"
#include "sctk_thread.h"
#include "sctk.h"
#include "sctk_mpcrun_client.h"

ssize_t
sctk_mpcserver_safe_read (int fd, void *buf, size_t count)
{
  char *tmp;
  ssize_t allready_readed = 0;
  ssize_t dcount = 0;

  tmp = buf;
  while (allready_readed < (ssize_t)count)
    {
      tmp += dcount;
      dcount = read (fd, tmp, count - allready_readed);
      if (!(dcount >= 0))
	{
	  perror ("safe_read");
	}
      assume (dcount >= 0);
      allready_readed += dcount;
    }
  assert ((ssize_t)count == allready_readed);
  return allready_readed;
}

ssize_t
sctk_mpcserver_safe_write (int fd, const void *buf, size_t count)
{
  ssize_t dcount;
  dcount = write (fd, buf, count);
  if (!(dcount == (ssize_t)count))
    {
      perror ("safe_write");
    }
  assume (dcount == (ssize_t)count);
  fsync (fd);
  return dcount;
}

int
sctk_tcp_connect_to (int portno, char *name)
{
  int clientsock_fd;
  struct sockaddr_in serv_addr;
  struct hostent *server;

  sctk_nodebug ("Try connection to %s on port %d", name, portno);

  clientsock_fd = socket (AF_INET, SOCK_STREAM, 0);
  if (clientsock_fd < 0)
    sctk_error ("ERROR opening socket");
  server = gethostbyname (name);
  if (server == NULL)
    {
      fprintf (stderr, "ERROR, no such host\n");
      exit (0);
    }
  memset ((char *) &serv_addr, 0, sizeof (serv_addr));
  serv_addr.sin_family = AF_INET;
  memmove ((char *) &serv_addr.sin_addr.s_addr,
	   (char *) server->h_addr, server->h_length);
  serv_addr.sin_port = htons ((unsigned short) portno);
  if (connect
      (clientsock_fd, (struct sockaddr *) (&serv_addr),
       sizeof (serv_addr)) < 0)
    {
      perror ("ERROR connecting");
      abort ();
    }

  sctk_nodebug ("Try connection to %s on port %d done", name, portno);
  return clientsock_fd;
}

int
sctk_mpcrun_client (char *request, void *in, size_t size_in,
		    void *out, size_t size_out)
{
  int fd;
  char req[MPC_ACTION_SIZE];

  memset (req, 0, MPC_ACTION_SIZE);
  assume (strlen (request) + 1 <= MPC_ACTION_SIZE);
  memcpy (req, request, strlen (request) + 1);

  fd = sctk_tcp_connect_to (sctk_mpcrun_client_port, sctk_mpcrun_client_host);
  sctk_nodebug ("Connected to host with req %s", req);
  sctk_mpcserver_safe_write (fd, req, MPC_ACTION_SIZE);
  sctk_mpcserver_safe_write (fd, &sctk_process_rank, sizeof (int));
  if (out != NULL)
    {
      sctk_mpcserver_safe_write (fd, &size_out, sizeof (unsigned long));
      sctk_mpcserver_safe_write (fd, out, size_out);
    }
  if (in != NULL)
    {
      unsigned long size;
      sctk_mpcserver_safe_read (fd, &size, sizeof (unsigned long));
      if (size != 0)
	{
	  sctk_nodebug ("%lu %lu", size, size_in);
	  assume (size == size_in);
	  sctk_mpcserver_safe_read (fd, in, size_in);
	}
      else
	{
	  sctk_nodebug ("Request fail %s", request);
	  close (fd);
	  return 1;
	}
    }

  close (fd);
  return 0;
}

void
sctk_mpcrun_barrier ()
{
  int total_nb;
  int done = 0;
  total_nb = sctk_process_number;

  do
    {
      assume (sctk_mpcrun_client
	      (MPC_SERVER_BARRIER, &done, sizeof (int), &total_nb,
	       sizeof (int)) == 0);
      sched_yield ();
      sctk_nodebug ("done %d tot %d", done, total_nb);
    }
  while (done != total_nb);
}

static int* socket_graph;
#define HOSTNAME_SIZE 4096
#define PORT_SIZE 4096

static int sockfd;
static char port[PORT_SIZE];
static void
sctk_create_recv_socket ()
{
  int portno;
  struct sockaddr_in serv_addr;

  sockfd = socket (AF_INET, SOCK_STREAM, 0);
  if (sockfd < 0)
    sctk_error ("ERROR opening socket");

  memset ((char *) &serv_addr, 0, sizeof (serv_addr));

  portno = 1023;
  sctk_debug ("Looking for an available port\n");
  do
    {
      portno++;
      serv_addr.sin_family = AF_INET;
      serv_addr.sin_addr.s_addr = INADDR_ANY;
      serv_addr.sin_port = htons ((unsigned short) portno);

    }
  while (bind (sockfd, (struct sockaddr *) &serv_addr,
	       sizeof (serv_addr)) < 0);
  sctk_debug ("Looking for an available port found %d\n", portno);
  sprintf (port, "%d", portno);

  listen (sockfd, 10);
}

void
sctk_mpcrun_client_init ()
{
  char *tmp;
  int rank;
  int i;

  tmp = getenv (MPC_SERVER_PORT);
  assume (tmp);
  if (tmp)
    sctk_mpcrun_client_port = atoi (tmp);
  sctk_nodebug ("PORT %s", tmp);
  tmp = getenv (MPC_SERVER_HOST);
  assume (tmp);
  sctk_nodebug ("HOSTNAME %s", tmp);
  sctk_mpcrun_client_host = tmp;
  sctk_process_number = sctk_get_process_nb ();
  assume (sctk_mpcrun_client
	  (MPC_SERVER_GET_RANK, &rank, sizeof (int), &sctk_process_number,
	   sizeof (int)) == 0);

  sctk_process_rank = rank;
  socket_graph = malloc(sctk_process_number*sizeof(int));
  memset(socket_graph,-1,sctk_process_number*sizeof(int));

}

void
sctk_mpcrun_client_init_connect ()
{
  int rank;
  int i;
  char dist_host[HOSTNAME_SIZE+PORT_SIZE+10];
  char local_host[HOSTNAME_SIZE+PORT_SIZE+10];

  sctk_mpcrun_client_init ();
  rank = sctk_process_rank;

  sctk_create_recv_socket ();

  memset(local_host,'\0',(HOSTNAME_SIZE+PORT_SIZE+10)*sizeof(char));
  gethostname(local_host, HOSTNAME_SIZE);
  sctk_debug("use port %s",port);

  sprintf(local_host+HOSTNAME_SIZE,"%s",port);

  sctk_debug("SEND %s %s",local_host,local_host+HOSTNAME_SIZE);
  assume(sctk_mpcrun_client(MPC_SERVER_REGISTER_HOST,NULL,0,local_host,(HOSTNAME_SIZE+PORT_SIZE+10)*sizeof(char)) == 0);

  for(i = 0; i < sctk_process_number; i++){
    if(i != rank){
      if(i < rank){
	memset(dist_host,'\0',(HOSTNAME_SIZE+PORT_SIZE+10)*sizeof(char));
	sctk_debug("Wait Recv peer");
	while(sctk_mpcrun_client(MPC_SERVER_GET_HOST,dist_host,(HOSTNAME_SIZE+PORT_SIZE+10)*sizeof(char),&i,sizeof(int)) != 0){
	  sleep(1);
	}
	sctk_debug("%d socket %s port %s",i,dist_host,dist_host+HOSTNAME_SIZE);
	socket_graph[i] = sctk_tcp_connect_to(atoi(dist_host+HOSTNAME_SIZE),dist_host);
	sctk_mpcserver_safe_write (socket_graph[i], &rank, sizeof(int));
	sctk_debug("CONNECT %d socket %d",i,socket_graph[i]);
      } else {
	int fd;
	int remote_rank;
	unsigned int clilen;
	struct sockaddr_in cli_addr;
	sctk_debug("Wait for connection");
	fd = accept (sockfd, (struct sockaddr *) &cli_addr, &clilen);
	
	sctk_mpcserver_safe_read (fd, &remote_rank, sizeof(int));
	socket_graph[remote_rank] = fd;
	sctk_debug("ACCEPT %d socket %d",remote_rank,socket_graph[remote_rank]);
      }
    }
  }
}
